/* wmsorsen
 * Copyright (C) 2006-2015 Ralf Hoffmann.
 * You can contact me at: ralf@boomerangsworld.de
 *   or http://www.boomerangsworld.de/cms/dockapps/wmsorsen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * This program is based on wmsensormon, (C) Liverbugg
 */

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>
#include <sensors/sensors.h>

#ifdef SENSORS_API_VERSION
#    if SENSORS_API_VERSION < 0x400
#        define API_LMSENSORS2
#    else
#        define API_LMSENSORS3
#    endif
#else
#    define API_LMSENSORS2
#endif

#ifdef API_LMSENSORS2
#    include <sensors/chips.h>
#endif

#include <sensors/error.h>
#include <string.h>
#include <stdlib.h>

#include "rc.hh"
#include <getopt.h>
#include <iostream>
#include <list>
#include <sstream>
#include <cstring>

extern "C" {
#include "../wmgeneral/wmgeneral.h"
#include "../wmgeneral/misc.h"
}

/* some content in the master3.xpm file has been taken from wmcpufreq */
#include "wmsorsen_master3.xpm"
#include "wmsorsen_mask.xbm"

#define VERSION "0.10.2"

#ifdef API_LMSENSORS3
#    define CONFIG_FILE "/etc/sensors3.conf"
#else
#    define CONFIG_FILE "/etc/sensors.conf"
#endif

void usage();
void version();
void display_units( int smp, int unit, int offset, double val );
void display_values( int, int, int );
void display_text( const std::string &text, int src_offset,
                   int dest_x, int dest_y );
void DrawBar( int sx, int sy, int w, int h, float percent, int dx, int dy );

class SensorValue
{
public:
    typedef enum { UNKNOWN, FAN, TEMP, VOLT } type_t;

    SensorValue() : _chipset( NULL ),
                    _name( "" ),
                    _feature( 0 ),
                    _type( UNKNOWN )
    {
    }

    SensorValue( const sensors_chip_name *chipset,
                 const std::string &name,
                 int feature,
                 type_t type ) : _chipset( chipset ),
                                 _name( name ),
                                 _feature( feature ),
                                 _type( type )
    {
    }

    const sensors_chip_name *getChipset() const
    {
        return _chipset;
    }

    std::string getName() const
    {
        return _name;
    }

    int getFeature() const
    {
        return _feature;
    }

    type_t getType() const
    {
        return _type;
    }

    bool isValid() const
    {
        if ( _chipset != NULL )
            return true;
        return false;
    }
private:
    const sensors_chip_name *_chipset;
    std::string _name;
    int _feature;
    type_t _type;
};

std::list<SensorValue> features;

#define MAX_DISPLAYED 5

struct sensor_display {
    SensorValue feature;
    double value, old_value;
    double limit1, limit2;
    int nr;
    std::string custom_text;
} display_feature[MAX_DISPLAYED];

static std::string getChipsetDescr( const sensors_chip_name *chipset )
{
    if ( chipset == NULL ) return "NULL";
    
    std::string res = chipset->prefix;

#ifdef API_LMSENSORS3
    switch ( chipset->bus.type ) {
        case SENSORS_BUS_TYPE_ISA:
            res += "_isa";
            break;
        case SENSORS_BUS_TYPE_ANY:
            res += "_any";
            break;
        case SENSORS_BUS_TYPE_I2C:
            res += "_i2c";
            break;
        case SENSORS_BUS_TYPE_PCI:
            res += "_pci";
            break;
        default:
            res += "_unknown";
            break;
    }
#else
    switch ( chipset->bus ) {
        case SENSORS_CHIP_NAME_BUS_ISA:
            res += "_isa";
            break;
        case SENSORS_CHIP_NAME_BUS_ANY:
            res += "_any";
            break;
        case SENSORS_CHIP_NAME_BUS_ANY_I2C:
            res += "_any-i2c";
            break;
        case SENSORS_CHIP_NAME_BUS_DUMMY:
            res += "_dummy";
            break;
        case SENSORS_CHIP_NAME_BUS_PCI:
            res += "_pci";
            break;
        default:
            res += "_unknown";
            break;
    }
#endif

    std::stringstream s;
    s << "_" << chipset->addr;
    res += s.str();

    return res;
}

static void getSensors()
{
    const sensors_chip_name *chipset;  
    int chip_nr = 0;
    SensorValue::type_t f1 = SensorValue::UNKNOWN;

    do {
#ifdef API_LMSENSORS3
        chipset = sensors_get_detected_chips( NULL, &chip_nr );
#else
        chipset = sensors_get_detected_chips( &chip_nr );
#endif
    
        /* Is there a chip here? */
        if ( chipset == NULL ) {
            break;
        } else {
            int nr1 = 0, nr2 = 0;
      
#ifdef API_LMSENSORS3
            const sensors_feature *feature = NULL;
#else
            const sensors_feature_data *feature = NULL;
#endif

            do {

#ifdef API_LMSENSORS3
                feature = sensors_get_features( chipset, &nr1 );
#else
                feature = sensors_get_all_features( *chipset, &nr1, &nr2 );
#endif

                if ( feature != NULL ) {

#ifdef API_LMSENSORS3
                    const sensors_subfeature *subfeature = NULL;
                    nr2 = 0;
                    do {
                        subfeature = sensors_get_all_subfeatures( chipset, feature, &nr2 );
                        
                        if ( subfeature == NULL ) break;
                        
                        if ( feature->type == SENSORS_FEATURE_TEMP ) {
                            f1 = SensorValue::TEMP;
                        } else if ( feature->type == SENSORS_FEATURE_FAN ) {
                            f1 = SensorValue::FAN;
                        } else if ( feature->type == SENSORS_FEATURE_IN ) {
                            f1 = SensorValue::VOLT;
                        } else {
                            f1 = SensorValue::UNKNOWN;
                        }
                        
                        if ( f1 != SensorValue::UNKNOWN ) {
                            SensorValue new_f( chipset,
                                               getChipsetDescr( chipset ) + "_" + subfeature->name,
                                               subfeature->number,
                                               f1 );
                            
                            features.push_back( new_f );
                        }
                    } while ( subfeature != NULL );
#else
                    if ( strncmp( feature->name, "temp", 4 ) == 0 ) {
                        f1 = SensorValue::TEMP;
                    } else if ( strncmp( feature->name, "fan", 3 ) == 0 ) {
                        f1 = SensorValue::FAN;
                    } else if ( strncmp( feature->name, "in", 2 ) == 0 ) {
                        f1 = SensorValue::VOLT;
                    } else {
                        f1 = SensorValue::UNKNOWN;
                    }

                    if ( f1 != SensorValue::UNKNOWN ) {
                        SensorValue new_f( chipset,
                                           getChipsetDescr( chipset ) + "_" + feature->name,
                                           feature->number,
                                           f1 );

                        features.push_back( new_f );
                    }
#endif
                }
            } while ( feature != NULL );
        }
    } while ( chip_nr > 0 );
}

static double getSensorValue( int i )
{
    double v = 0.0;

    if ( i < 0 || i >= MAX_DISPLAYED ) return v;

    if ( display_feature[i].feature.isValid() == false )
        return v;

#ifdef API_LMSENSORS3
    if ( sensors_get_value( display_feature[i].feature.getChipset(), display_feature[i].feature.getFeature(), &v ) != 0 ) {
#else
    if ( sensors_get_feature( *display_feature[i].feature.getChipset(), display_feature[i].feature.getFeature(), &v ) != 0 ) {
#endif
        printf( "ERROR: Couldn't get value of %s\n", display_feature[i].feature.getName().c_str() );
    }
    display_feature[i].value = v;
    return v;
}

#if 0
static void getSensorValues()
{
    int i;

    for ( i = 0; i < MAX_DISPLAYED; i++ ) {
        if ( display_feature[i].feature.isValid() == false )
            break;
        printf( "Value for %s:%f\n", display_feature[i].feature.getName().c_str(), getSensorValue( i ) );
    }
}
#endif

static void printSensors()
{
    printf( "Available sensors (current value in parentheses):\n" );
    for ( std::list<SensorValue>::const_iterator it1 = features.begin();
          it1 != features.end();
          it1++ ) {
        double v = 0.0;

#ifdef API_LMSENSORS3
        sensors_get_value( it1->getChipset(), it1->getFeature(), &v );
#else
        sensors_get_feature( *( it1->getChipset() ), it1->getFeature(), &v );
#endif

        if ( (*it1).getType() == SensorValue::FAN ) {
            printf( "FAN:%s (%.2f)\n", (*it1).getName().c_str(), v );
        } else if ( (*it1).getType() == SensorValue::TEMP ) {
            printf( "TEMP:%s (%.2f)\n", (*it1).getName().c_str(), v );
        } else if ( (*it1).getType() == SensorValue::VOLT ) {
            printf( "Voltage:%s (%.2f)\n", (*it1).getName().c_str(), v );
        }
    }
}

static SensorValue findFeature( const std::string name )
{
    std::list<SensorValue>::const_iterator it1;
  
    for ( it1 = features.begin();
          it1 != features.end();
          it1++ ) {
        if ( (*it1).getName() == name ) {
            break;
        }
    }
    if ( it1 != features.end() ) {
        return *it1;
    }
    return SensorValue();
}

int main( int argc, char **argv )
{
    long delay = 100000;
    FILE *config_file;
    int unit = 0;
    int offset = 0, offset2 = 1, i = 0;
    //   float volt0_per;

    // init libsensors
    if ( ( config_file = fopen( CONFIG_FILE, "r" ) ) == NULL ) {
        fprintf( stderr,
                 "\tCan't find \"%s\".\n"           \
                 "\tCheck the instalation of lm_sensors.\n", CONFIG_FILE );
        exit( EXIT_FAILURE );
    } else {
        if ( sensors_init( config_file ) ) {
            fprintf( stderr, "\tProblems with \"%s\".\n", CONFIG_FILE );
            exit( EXIT_FAILURE );
        }
    }   
    fclose( config_file );

    // get all available sensors
    getSensors();

    // read config file
    config_arg_map rccontent;
    std::string filename;

    if ( getenv( "HOME" ) != NULL ) {
        filename = getenv( "HOME" );
        filename += "/.wmsorsenrc";
    }
    readConfig( filename, rccontent );

    // check display values
    int temp_count = 0;
    int fan_count = 0;
    for ( i = 0; i < MAX_DISPLAYED; i++ ) {
        std::string keyname;
        std::stringstream sts;
     
        sts << "monitor";
        sts << ( i + 1 );
        keyname = sts.str();
        if ( rccontent.count( keyname ) > 0 ) {
            std::vector<std::string> a = rccontent[keyname];
            std::string sensorname, tstr;
            double limit1val = -1.0, limit2val = -1.0;

            // sensor name found
            if ( a.size() > 0 ) {
                sensorname = a[0];
            }

            // limit1 found
            if ( a.size() > 1 ) {
                char *rest = NULL;
         
                tstr = a[1];
                limit1val = strtod( tstr.c_str(), &rest );
                if ( limit1val == 0.0 && rest == tstr.c_str() )
                    limit1val = -1.0;
            }

            // limit2 found
            if ( a.size() > 2 ) {
                char *rest = NULL;
         
                tstr = a[2];
                limit2val = strtod( tstr.c_str(), &rest );
                if ( limit2val == 0.0 && rest == tstr.c_str() )
                    limit2val = -1.0;
            }

            if ( sensorname.length() > 0 ) {
                // now register this value
                display_feature[i].feature = findFeature( sensorname );
                if ( display_feature[i].feature.isValid() == false ) {
                    std::cout << "Given sensor name " << sensorname << " was not found." << std::endl;
                } else {
                    display_feature[i].limit1 = limit1val;
                    display_feature[i].limit2 = limit2val;

                    if ( display_feature[i].feature.getType() == SensorValue::FAN ) {
                        display_feature[i].nr = fan_count++;
                    } else if ( display_feature[i].feature.getType() == SensorValue::TEMP ) {
                        display_feature[i].nr = temp_count++;
                    } else {
                        display_feature[i].nr = 0;
                    }

                    display_feature[i].custom_text = "";

                    std::stringstream stt;
                    stt << "text";
                    stt << ( i + 1 );
                    std::string textname = stt.str();
                    if ( rccontent.count( textname ) > 0 ) {
                        std::vector<std::string> textcontent = rccontent[textname];
                        if ( ! textcontent.empty() ) {
                            display_feature[i].custom_text = textcontent[0];
                        }
                    }
                }
            }
        }
    }

    // parse cli arguments
    while ( 1 ) {
        int option_index, c;
        static struct option long_options[] = {
            { "help", no_argument, NULL, 'h' },
            { "print", no_argument, NULL, 'p' },
            { "version", no_argument, NULL, 'V' },
            { "refresh", required_argument, NULL, 'r' },
            { "fahrenheit", no_argument, NULL, 'f' },
            { NULL, 0, NULL, 0 }
        };

        c = getopt_long( argc, argv, "hpVr:f", long_options, &option_index );
        if ( c == -1 ) break;

        switch( c ) {
            case 'h':
                usage();
                exit( EXIT_SUCCESS );
                break;
            case 'V':
                version();
                exit( EXIT_SUCCESS );
                break;
            case 'p':
                printSensors();
                exit( EXIT_SUCCESS );
                break;
            case 'r':
                delay = atol( optarg );
                break;
            case 'f':
                unit = 1;
                break;
            default:
                break;
        }
    }

    char *wmsorsen_mask_bits_copy = (char*)malloc( sizeof( char ) *
                                                   sizeof( wmsorsen_mask_bits ) /
                                                   sizeof( wmsorsen_mask_bits[0] ) );
    memcpy( wmsorsen_mask_bits_copy, wmsorsen_mask_bits,
            sizeof( wmsorsen_mask_bits ) / sizeof( wmsorsen_mask_bits[0] ) );

    char **wmsorsen_master3_xpm_copy = (char**)malloc( sizeof( char* ) *
                                                       sizeof( wmsorsen_master3_xpm ) /
                                                       sizeof( wmsorsen_master3_xpm[0] ) );
    for ( unsigned int ti1 = 0;
          ti1 < sizeof( wmsorsen_master3_xpm ) /
              sizeof( wmsorsen_master3_xpm[0] );
          ti1++ ) {
        wmsorsen_master3_xpm_copy[ti1] = strdup( wmsorsen_master3_xpm[ti1] );
    }

    openXwindow( argc, argv,
                 wmsorsen_master3_xpm_copy, wmsorsen_mask_bits_copy,
                 wmsorsen_mask_width, wmsorsen_mask_height );

    for (;;) {
        int yoffset = -10;
     
        for ( i = 0; i < MAX_DISPLAYED; i++ ) {
            yoffset += 14;
            getSensorValue( i );
            if ( display_feature[i].value != display_feature[i].old_value ) {
                display_feature[i].old_value = display_feature[i].value;
         
                if ( i < 4 ) {
                    // text area
           
                    offset = 0;
                    offset2 = 0;
                    if ( display_feature[i].feature.getType() == SensorValue::FAN ) {
                        if ( display_feature[i].limit1 >= 0.0 &&
                             display_feature[i].value < display_feature[i].limit1 ) {
                            offset = 7;
                            offset2 = 7;
                        }
                    } else if ( display_feature[i].feature.getType() == SensorValue::TEMP ) {
                        if ( display_feature[i].limit2 >= 0.0 &&
                             display_feature[i].value > display_feature[i].limit2 ) {
                            offset = 7;
                            offset2 = 7;
                        } else if ( display_feature[i].limit1 >= 0.0 &&
                                    display_feature[i].value > display_feature[i].limit1 ) {
                            offset = 14;
                            offset2 = 14;
                        }
                    } else if ( display_feature[i].feature.getType() == SensorValue::VOLT ) {
                        if ( display_feature[i].limit1 >= 0.0 &&
                             display_feature[i].limit2 >= 0.0 &&
                             ( display_feature[i].value > display_feature[i].limit2 ||
                               display_feature[i].value < display_feature[i].limit1 ) ) {
                            offset = 7;
                            offset2 = 7;
                        }
                    }

                    if ( ! display_feature[i].custom_text.empty() ) {
                        display_text( display_feature[i].custom_text,
                                      offset, 5, yoffset );
                    } else {
                        if ( display_feature[i].feature.getType() == SensorValue::FAN ) {
                            copyXPMArea( 46, 87 + offset, 23, 7, 5, yoffset );
                            copyXPMArea( ( display_feature[i].nr + 1 ) * 5, 65 + offset, 5, 7, 23, yoffset );
                        } else if ( display_feature[i].feature.getType() == SensorValue::VOLT ) {
                            //TODO needs some work
                        } else {
                            copyXPMArea( 0, 87 + offset, 23, 7, 5, yoffset );
                            //copyXPMArea( 69, 87 + offset, 5, 7, 23, yoffset );
                            copyXPMArea( ( display_feature[i].nr + 1 ) * 5, 65 + offset, 5, 7, 23, yoffset );
                        }
                    }
           
                    if ( display_feature[i].feature.getType() == SensorValue::FAN ) {
                        display_values( (int)display_feature[i].value, yoffset - 4, offset2 );
                    } else if ( display_feature[i].feature.getType() == SensorValue::VOLT ) {
                        //TODO needs some work
                    } else {
                        double val = display_feature[i].value;

                        if ( unit == 1 ) {
                            val *= 1.8;
                            val += 32;
                        }
   
                        display_values( (int)( val * 10.0 ), yoffset - 4, offset2);
                        display_units( unit, offset, yoffset - 4, val );
                    }
                } else {
                    float volt_per = -1.0;
                    if ( display_feature[i].feature.getType() == SensorValue::FAN ) {
                        //TODO
                    } else if ( display_feature[i].feature.getType() == SensorValue::TEMP ) {
                        //TODO
                    } else if ( display_feature[i].feature.getType() == SensorValue::VOLT ) {
                        if ( display_feature[i].limit1 > 0.0 &&
                             display_feature[i].limit2 > 0.0 ) {
                            volt_per = display_feature[i].value - display_feature[i].limit1;
                            volt_per /= display_feature[i].limit2 - display_feature[i].limit1;
                            volt_per *= 100.0;
                        }
                    }
                    DrawBar(0, 109, 58, 4, volt_per, 3, 58);
                }
            }
        }
        RedrawWindow();
        usleep( delay );
    }
}  

void display_bar( int sx, int sy, int w, int h, float percent, int dx, int dy )
{
    static int tx;
    int temp = sx ;
    tx = (int)( (float)w * ( (float)percent / (float)100.0 ) );
  
    while ( temp <= tx ) {
        copyXPMArea( temp, sy, 1, h, dx, dy );
        temp++;
    }
  
    copyXPMArea( sx, sy, tx, h, dx, dy );
    copyXPMArea( sx, sy + h + 3, w - tx, h, dx + tx, dy );
}

void DrawBar( int sx, int sy, int w, int h, float percent, int dx, int dy )
{
    static int tx;
    tx = (int)( (float)w * ( (float)percent / (float)100.0 ) );
    if ( tx > w ) tx = w;
  
    copyXPMArea( sx, sy, tx, h, dx, dy );
    copyXPMArea( sx, sy + h + 3, w - tx, h, dx + tx, dy );
}


void display_units( int unit, int offset, int offset2, double val )
{
    if ( unit == 0 ) {
        copyXPMArea( 66, 65 + offset, 6, 7, 53, 4 + offset2 );
    } else {
        copyXPMArea( 73, 65 + offset, 6, 7, 53, 4 + offset2 );
    }
    if ( val < 100.0 ) {
        copyXPMArea( 51, 70 + offset, 2, 2, 45, 10 + offset2 );
    } else {
        copyXPMArea( 54, 70 + offset, 2, 2, 45, 10 + offset2 );
    }
}

void display_values( int temp, int offset, int offset2 )
{
    char text[5], num1, num2, num3, num4;
  
    sprintf( text, "%03d", temp );
    num1 = text[0] - '0';
    num2 = text[1] - '0';
    num3 = text[2] - '0';
    num4 = text[3] - '0';

    if ( num1 )
        copyXPMArea( 5 * num1, 65 + offset2, 5, 7, 33, 4 + offset );
    else
        copyXPMArea( 60, 65 + offset2, 5, 7, 33, 4 + offset );
    copyXPMArea( 5 * num2, 65 + offset2, 5, 7, 39 + 1, 4 + offset );
    copyXPMArea( 5 * num3, 65 + offset2, 5, 7, 45 + 2, 4 + offset );
    copyXPMArea( 5 * num4, 65 + offset2, 5, 7, 51 + 3, 4 + offset );
}

void display_text( const std::string &text, int src_offset,
                   int dest_x, int dest_y )
{
    int chars = text.length();
    int char_width = 6;

    chars = ( chars > 4 ) ? 4 : chars;

    for ( int i = 0; i < chars; i++ ) {
        int base_x = 1;
        char ch = std::tolower( text[i] );
        
        if ( ch >= 'a' && ch <= 'z' ) {
            copyXPMArea( base_x +  ( ch - 'a' ) * char_width, 123 + src_offset,
                         char_width, 7, dest_x, dest_y );
        } else if ( ch >= '0' && ch <= '9' ) {
            copyXPMArea( ( ch - '0' ) * 5, 65 + src_offset, 5, 7, dest_x, dest_y );
        }

        dest_x += char_width;
    }
}

void usage()
{
    puts( "\nwmsorsen "VERSION" - window maker dockapp for motherboard sensors.\n"
          "Options : -h, --help            this help screen\n"
          "          -V, --version         print version information\n"
          "          -p, --print           print all available sensor information\n"
          "          -f, --fahrenheit      show temperatures in Fahrenheit\n"
          "          -r, --refresh         refresh time (in microseconds, default 100000)\n\n"
          "Also check $HOME/.wmsorsenrc for configuration\n\n"
          "Ralf Hoffmann <ralf@boomerangsworld.de>\nHomepage: http://www.boomerangsworld.de/cms/dockapps/wmsorsen/\n"
          );
    exit( EXIT_SUCCESS );
}

void version()
{
    printf( "\nwmsorsen "VERSION"\n"
            "Ralf Hoffmann <ralf@boomerangsworld.de>\n"
            "http://www.boomerangsworld.de/cms/dockapps/wmsorsen/\n" );
    exit( EXIT_SUCCESS );
}  
